# Test Specification

The RPCD tests are written according to its feature requirements, defined in the
[function specification](./functionspec.md#requirements) and is tested through
unit tests.

No functional, or functional-API tests are performed, as the functionality
exposed through ubus can be tested by invoking the callbacks themselves, using
the compiled shared library of the rpcd source files. RPCD does not depend
on any external ubus objects, daemons or processes to be running to fulfill its
functionality.

## Prerequisites

Apart from the required library dependencies specified in the
[README](../README.md#dependencies) the test environment has to be prepared with
an rpcd configuration file, dummy configuration files to read and write to,
access lists with `uci_granular` options set, and users. These files may be
found at `./test/cmocka/files/` and are appropriately setup to a test
environment by the gitlab-ci pipeline setup script `./gitlab-ci/setup.sh`.

## Unit Tests

The RPCD unit tests are written in Cmocka, and consists of one test suite, the
granular access list test suite. The tests are written as traditional unit
tests, invoking indiviudal functions from the shared library compiled from the
source files of rpcd.

### Granular Access Control Tests

The granular access control tests are desgined to cover all possible use cases
of access control over UCI, exposed by the RPCD ubus API.

The methods tested are primarily the UCI object methods:
1. `get`
2. `set`
3. `add`
4. `delete`
5. `rename`
6. `commit`

For all ubus the methods the tests verify several input variations, ensuring
that the access lists are properly applied for all test cases, i.e. by providing
the input arguments, `type`, `name`, `option` in different valid
combinations. All methods have a false positive test case to verify that

Furthermore, the tests verify that prior functionality is intact (no ubusrpc
session provided grants full access, no access list restriction grants access),
that new functionality is in place and functions according to the requirements.

#### Test Structures

Each ubus method callback has a wrapper prepared in `unit_tests_gacl.c`, for a
more convenient way to prepare the input.

```
static int perform_uci_rename(char *user, char *cfg, char *section, char *option, char *name, char *id);
static int perform_uci_delete(char *cfg, char *section, char *type, char *option, char *options, char *id);
static int perform_uci_set(char *user, char *cfg, char *section, char *type, char *values, char *id);
static int perform_uci_add(char *user, char *cfg, char *section, char *type, char *values, char *id);
static struct blob_attr *perform_uci_get(char *user, char *cfg, char *section, char *type, char *option, char *id);
```

The wrappers takes arguments such as configuration file, section, type,
ubus_rpc_session etc. and prepares the `blob_buf` structure accordingly, then
invokes the source file callback. To validate the test, multiple methods are
used test types:

1. Validate the return value of the ubus call, these represents a ubus return
status from libubus, i.e. `UBUS_STATUS_OK`, `UBUS_STATUS_PERMISSION_DENIED`.
2. When applicable, validate that the ubus output passed to `ubus_send_reply(3)`
reflects the expected output.
3. Validate by invoking uci get on the affected configuration and validate the
json against pre-defined expected output.

#### Test Case Table

| Serial ID		| Test Case Name	                                                            			|
| :---			| :---				                                                            			|
| 1				| [test_settings_name_access](#test_settings_name_access)	                                |
| 2				| [test_granular_access_get_by_type](#test_granular_access_get_by_type)	                    |
| 3				| [test_granular_access_get_by_type_fail](#test_granular_access_get_by_type_fail)	        |
| 4				| [test_uci_granular_parse_match_definition](#test_uci_granular_parse_match_definition)     |
| 5				| [test_granular_access_get_by_config](#test_granular_access_get_by_config)	                |
| 6				| [test_granular_access_get_by_option](#test_granular_access_get_by_option)	                |
| 7				| [test_granular_access_get_by_option_fail](#test_granular_access_get_by_option_fail)       |
| 8				| [test_granular_multi_session](#test_granular_multi_session)	                			|
| 9				| [test_granular_access_get_by_section](#test_granular_access_get_by_section)	            |
| 10			| [test_granular_access_get_by_section_fail](#test_granular_access_get_by_section_fail)     |
| 11			| [test_granular_access_set_by_section](#test_granular_access_set_by_section)	            |
| 12			| [test_granular_access_set_by_section_fail](#test_granular_access_set_by_section_fail)     |
| 13			| [test_granular_add](#test_granular_add)	                                				|
| 14			| [test_granular_add_fail](#test_granular_add_fail)	                						|
| 15			| [test_granular_delete_by_section](#test_granular_delete_by_section)	                    |
| 16			| [test_granular_delete_by_section_fail](#test_granular_delete_by_section_fail)	            |
| 17			| [test_granular_delete_by_type](#test_granular_delete_by_type)	                			|
| 18			| [test_granular_delete_by_type_fail](#test_granular_delete_by_type_fail)	            	|
| 19			| [test_granular_rename_option](#test_granular_rename_option)	                			|
| 20			| [test_granular_rename_section](#test_granular_rename_section)	            				|
| 21			| [test_granular_rename_option_fail](#test_granular_rename_option_fail)	                    |
| 22			| [test_granular_rename_section_fail](#test_granular_rename_section_fail)	                |

#### test_settings_name_access

##### Description

Load the granular access control restrictions for the user `user`, configuration
file `test`, of section `settings` to verify that the expected access list
section is loaded.

##### Test Steps

Load the desired access list configuration through `rpc_uci_load_granular(5)`,
use `rpc_uci_check_granular_is_allowed(1)`.

##### Test Expected Results

The expected result is for `radio` and `band` to return true as allowed options,
while no others should.

#### test_granular_access_get_by_type

##### Description

Invoke the ubus callback for `uci->get` by only
providing `section` and `ubus_rpc_session` as input, programmatically, to verify
that access lists returns correct output and applies access list for multiple
output sections.

##### Test Steps

Use the wrapper `perform_uci_get(6)` and verify that the right amount of
sections are returned and that the returned values are of the expected values.

##### Test Expected Results

The expected results that there are a total of three options are returned,
contained within two sections and the return value from `perform_uci_get(6)`
should be 0, `UBUS_STATUS_OK`. Verify for the section named `settings` that the
values `radio` is returned as `wl1` and `band` is set to `b`.

#### test_granular_access_get_by_type_fail

##### Description

Invoke the ubus callback for `uci->get` by only
providing `rule` and `ubus_rpc_session` as input, programmatically, to verify
that access lists does not return any output, as this section is not explicitly
allowed by the granular access list for `user`.

##### Test Steps

Use the wrapper `perform_uci_get(6)` and verify the number of returned sections
and options.

##### Test Expected Results

The expected results is that there are no options, or sections, in the prepared
ubus output.

#### test_uci_granular_parse_match_definition

##### Description

Regression test to verify that providing NULL as input does not cause
segmentation fault on `uci_granular_parse_match_definition(2)`.

##### Test Steps

Call the method `uci_granular_parse_match_definition(2)`, providing both
arguments as NULL, and providing an empty `struct blob_attr` as first argument
with NULL as second argument.

##### Test Expected Results

The expected result is for no segmentation fault or memory leak to be caused.

#### test_granular_access_get_by_config

##### Description

Invoke the ubus callback for `uci->get`, providing
only the configuration file and `ubus_rpc_session` as input, verifying that
access lists are applied when reading an entire configuration file.

##### Test Steps

Use the wrapper `perform_uci_get(6)` and verify that the right amount of
sections are returned and that the returned values are of the expected values.

##### Test Expected Results

The expected results is that there are a total of four options returned,
contained within three sections and the return value from `perform_uci_get(6)
should be 0, `UBUS_STATUS_OK`.. Use libjson-editor to verify for the section
named `settings` that the values `radio` is returned as `wl1` and `band` is set
to `b`.

#### test_granular_access_get_by_option

##### Description

Invoke the ubus callback for `uci->get`, explicitly
providing every argument as input, reading the option `test.settings.radio`,
verifying that access lists are applied when reading just one option.

##### Test Steps

Use the wrapper `perform_uci_get(6)` and verify that the expected amount of
values and keys returned.

##### Test Expected Results

The expected result is for the value `wl1` to be returned in the ubus output,
and the return value from `perform_uci_get(6)` should be 0, `UBUS_STATUS_OK`.

#### test_granular_access_get_by_option_fail

##### Description

Invoke the ubus callback for `uci->get`, explicitly
providing every argument as input, attempting to read an option which is not
whistlisted by the access list, verifying that access lists are applied when
reading just one option, and in this test case, is denied.

##### Test Steps

Use the wrapper `perform_uci_get(6)` and verify the returned ubus output.

##### Test Expected Results

No keys should be returned, and the return value of `perform_uci_get(6)` should
be 6, `UBUS_STATUS_PERMISSION_DENIED`.

#### test_granular_multi_session

##### Description

Login to a second session simultaneuously, using the user `admin`, to obtain
another `ubus_rpc_session`. Invoke `uci->get` programmatically with both IDs to
ensure that the granular access lists support multiple sessins to be active.

##### Test Steps

Create a new sessions using the wrapper `create_new_session(1)`, perform an
invoke for `uci->get` using the newly obtained session (using the `admin` user),
then perform one with the normal test context (using the `user` user). Verify
the returned values, make sure they adhere to the corresponding access lists.

##### Test Expected Results

Both users should experience restricted access, albeit to different options.
Out of the tested values admin should have access to `test.settings.type` but
not `test.settings.channel`, while `user` should have access to
`test.settings.radio` and not `test.settings.channel`, correspondingly,
the return values should be `UBUS_STATUS_OK` and
`UBUS_STATUS_PERMISSION_DENIED`.

#### test_granular_access_get_by_section

##### Description

Invoke the ubus callback for `uci->get`, providing
the configuration file, section name and `ubus_rpc_session` as input to verify
that access lists are applied when reading a section.

##### Test Steps

Use the wrapper `perform_uci_get(6)` and verify that the expected amount of
options, and their values are returned.

##### Test Expected Results

The expected results is that there are two options returned, `radio` set to
`wl1` and `band` set to `b` and the return value from`perform_uci_get(6)` to be
0, `UBUS_STATUS_OK`.

#### test_granular_access_get_by_section_fail

##### Description

Invoke the ubus callback for `uci->get`, providing
the configuration file, section name and `ubus_rpc_session` to verify that
access is not granted to a options within a section not listed in the granular
access list.

##### Test Steps

Use the wrapper `perform_uci_get(6)` to invoke `uci->get` and verify that no
values are returned.

##### Test Expected Results

The expected results is that no values are returned.

#### test_granular_access_set_by_section

##### Description

Invoke the ubus callback for `uci->set`, attempting
to set an already existing value in a section with partial access, and setting
an entirely new value in a section with unrestricted access.

##### Test Steps

Use the wrapper `perform_uci_set(6)`, invoking `uci->set`, in order to set a
value in the configuration file, use the function `rpc_uci_revert_commit(3)` to
commit the change, the second argument should be a `struct blob_buf` holding the
configuration to commit, passing the third argument, `bool commit`, as true in
order to trigger a commit. The results may be verified by reading the
configuration file using the wrapper `perform_uci_get(6)`.

##### Test Expected Results

The expected result is for the option `test.settings.radio` has been set to
`new_radio` and for the option`test.meta.name` has been set to `test`.

#### test_granular_access_set_by_section_fail

##### Description

Invoke the ubus callback for `uci->set`, attempting
set an existing option to which access is not granted to, and attempting to add
option to a section with restricted access.

##### Test Steps

Use the wrapper `perform_uci_set(6)`, invoking `uci->set`, in order to attempt
to set value in the configuration file, use the function
`rpc_uci_revert_commit(3)` to commit the change, the second argument should be a
`struct blob_buf` holding the configuration to commit, passing the third
argument, `bool commit`, as true in order to trigger a commit. The results may
be verified by reading the configuration file using the wrapper
`perform_uci_get(6)`.

##### Test Expected Results

The expected result is for no option to set or added.

#### test_granular_add

##### Description

Invoke the ubus callback for `uci->add`, attempting
add an entirely new section, and using it to set values in a preexisting
section.

##### Test Steps

Use the wrapper `perform_uci_add(6)`, invoking `uci->add`, in order to add a
section or value and `rpc_uci_revert_commit(3)` to commit the change, the second
argument should be a `struct blob_buf` holding the configuration to commit, passing the third
argument, `bool commit`, as true in order to trigger a commit. The results may
be verified by reading the configuration file using the wrapper
`perform_uci_get(6)`.

##### Test Expected Results

The expected result is for the section:
```
config meta 'new'
    option completely 'new'
    option additional 'add'
    option additional2 'add2'
    option additional3 'add3'
```

to be added to the configuration file.

#### test_granular_add_fail

##### Description

Invoke the ubus callback for `uci->add`, attempting
add an entirely new section, adding options to an existing section with some
access, and adding to an existing section with no access.

##### Test Steps

Use the wrapper `perform_uci_add(6)`, invoking `uci->add`, in order to add a
section or value and `rpc_uci_revert_commit(3)` to commit the change, the second
argument should be a `struct blob_buf` holding the configuration to commit,
passing the third argument, `bool commit`, as true in order to trigger a commit.
The results may be verified by reading the configuration file using the wrapper
`perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected results is for no new sections or options to be added and for the
return value from the wrapper `perform_uci_add(6)` to be 6,
`UBUS_STATUS_PERMISSION_DENIED`.

#### test_granular_delete_by_section

##### Description

Invoke the ubus callback for `uci->delete`, deleinge options and
entire sections, to which unrestricted access is granted.

##### Test Steps

Use the wrapper `perform_uci_delete(6)`, invoking `uci->delete`, in order to
delete a section or value and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected results is for individual options from the section `new` to be
deleted, and eventually the entire section, additionally, for the return value
from `perform_uci_delete(6)` to be 0, `UBUS_STATUS_OK`.

#### test_granular_delete_by_section_fail

##### Description

Invoke the ubus callback for `uci->delete`, attempting delete options and
entire sections, to which limited, or no access, is granted.

##### Test Steps

Use the wrapper `perform_uci_delete(6)`, invoking `uci->delete`, in order to
delete a section or value and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected results is for no sections or options to be deleted and for the
return value from the wrapper `perform_uci_add(6)` to be 6,
`UBUS_STATUS_PERMISSION_DENIED`.

#### test_granular_delete_by_type

##### Description

Invoke the ubus callback for `uci->delete`, attempting delete options and
entire sections, only specifying type as input, to which full access is granted.

##### Test Steps

Use the wrapper `perform_uci_delete(6)`, invoking `uci->delete`, in order to
delete a section or value and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected results is for individual options from the `new` section to be
deleted during the test, and finally the entire section, along with the `test`
section.

#### test_granular_delete_by_type

##### Description

Invoke the ubus callback for `uci->delete`, attempting delete options and
entire sections, only specifying type as input, to which limited, or no access,
is granted.

##### Test Steps

Use the wrapper `perform_uci_delete(6)`, invoking `uci->delete`, in order to
delete a section or value and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected result is for the return code 6, `UBUS_STATUS_PERMISSION_DENIED` to
be returned for each and every call to the wrapper `perform_uci_delete(6)`,
additionally verified via `perform_uci_get(6)` that no options were deleted.

#### test_granular_rename_option

##### Description

Invoke the ubus callback for `uci->rename`, attempting rename an option in a
section to which full access is granted.

##### Test Steps

Use the wrapper `perform_uci_rename(6)`, invoking `uci->rename`, in order to
rename an option, and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected result is for the return code 0, `UBUS_STATUS_OK` to
be returned, additionally verified via `perform_uci_get(6)` that the option
`test.test.name` is no longer in the configuration, but `test.test.rename` is.

#### test_granular_rename_section

##### Description

Invoke the ubus callback for `uci->rename`, attempting rename a section to which
full access is grantd.

##### Test Steps

Use the wrapper `perform_uci_rename(6)`, invoking `uci->rename`, in order to
rename a section, and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected result is for the return code 0, `UBUS_STATUS_OK` to be returned,
additionally verified via `perform_uci_get(6)` that the section
`test.test` is no longer in the configuration, but `test.newer` is.


#### test_granular_rename_option_fail

##### Description

Invoke the ubus callback for `uci->rename`, attempting rename an option in a
section to which full access is not granted.

##### Test Steps

Use the wrapper `perform_uci_rename(6)`, invoking `uci->rename`, in order to
rename an option, and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected result is for the return code 6, `UBUS_STATUS_PERMISSION_DENIED` to
be returned, additionally verified via `perform_uci_get(6)` that the option
`test.settings.new_radio` is not in the configuration, but `test.settings.radio`
is.

#### test_granular_rename_section_fail

##### Description

Invoke the ubus callback for `uci->rename`, attempting rename a section to which
full access is not granted.

##### Test Steps

Use the wrapper `perform_uci_rename(6)`, invoking `uci->rename`, in order to
rename a section, and `rpc_uci_revert_commit(3)` to commit the change,
the second argument should be a `struct blob_buf` holding the configuration to
commit, passing the third argument, `bool commit`, as true in order to trigger a
commit. The results may be verified by reading the configuration file using the
wrapper `perform_uci_get(6)`, and verifying the return values.

##### Test Expected Results

The expected result is for the return code 6, `UBUS_STATUS_PERMISSION_DENIED` to
be returned, additionally verified via `perform_uci_get(6)` that the section
`test.settings` is still in the configuration, and no section
`test.new_settings` is present.

## Writing New Tests

When writing new tests, use one of the readily available wrappers for adding,
setting, deleting, renaming and getting. Currently no other functions are
affacted by granular access control, meaning no additional wrappers are
currently necessary, however, should this change, a new wrapper should be
implemented in such a way that it accepts all the possible arguments to form
any kind of ubus invoke query, and with the possibility to leave out optional
arguments.

The response is most easily verified by passing the return value of the invoke,
and verifying the actual changes to the configuration through the wrapper
`perform_uci_get(6)` on the affected configuration file, section, type etc.

Note: on an add, delete, set and rename, it is necessary to commit the changes
to the configuratin file through `rpc_uci_revert_commit(3)`, the second argument
should be a `struct blob_buf` holding the configuration to commit, passing the
third argument, `bool commit`, as true in order to trigger a commit.

```
blob_buf_init(&bb, 0);
blobmsg_add_string(&bb, "config", "test");
rpc_uci_revert_commit(NULL, bb.head, true);
```

The UCI configuration used for testing is the `test` configuration, found at
`/etc/config/test`, which is the only one with granular restrictions by the
`admin.json` and `user.json` access lists, if some new requirement is formed,
or some test has to be added, either of the access lists may be modified,
however, must of the tests are performed by the `user` login, affected by
`user.json`. Alternatively, another access list may be created and added
with `list read` or `list write`, depending on the structure of the access list
file, as all restrictions are gathered and applied as a union.
