/*
 * RPC login guard mechanism
 *
 * Copyright (C) 2019 IOPSYS
 *
 * Author: Matija Amidzic <matija.amidzic@sartura.hr>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <syslog.h>
#include <libubox/avl.h>
#include <libubox/avl-cmp.h>
#include <libubox/uloop.h>
#include <uci.h>
#include <string.h>
#include <stdlib.h>

#include <rpcd/login_guard.h>

#define CONFIG_GUARD "brute-force"
#define CONFIG_MAX_ATTEMPTS "max_incorrect_attempts"
#define CONFIG_MAX_TIME_BEFORE_LOCK "max_time_before_locking"
#define CONFIG_LOCK_DURATION "account_lock_duration"

struct login_guard_s {
	struct avl_tree login_attempts;

	int max_attempts;
	int max_time_before_lock;
	int lock_duration;

	bool enabled;
};

struct login_attempt_ctx {
	struct avl_node avl;

	struct uloop_timeout lock_timer;
	struct uloop_timeout unlock_timer;

	bool is_locked;
	int incorrect_attempts;
};
static struct login_attempt_ctx *login_guard_find_attempt(const char *username);
static void login_guard_lock_timeout(struct uloop_timeout *t);
static void login_guard_unlock_timeout(struct uloop_timeout *t);
static struct login_attempt_ctx *login_guard_create_attempt(const char *username);
static void login_guard_delete_attempt(struct login_attempt_ctx *login_attempt);

static struct login_guard_s login_guard = {0};

static struct login_attempt_ctx *login_guard_find_attempt(const char *username)
{
	struct login_attempt_ctx *attempt = NULL;
	struct login_attempt_ctx *temp = NULL;
	avl_for_each_element(&login_guard.login_attempts, temp, avl) {
		if (strcmp(temp->avl.key, username) == 0) {
			attempt = temp;
			// we found the attempt, break
			break;
		}
	}
	return attempt;
}

static void login_guard_lock_timeout(struct uloop_timeout *t)
{
	struct login_attempt_ctx *login_attempt;

	login_attempt = container_of(t, struct login_attempt_ctx, lock_timer);
	login_attempt->incorrect_attempts = 0;
}

static void login_guard_unlock_timeout(struct uloop_timeout *t)
{
	struct login_attempt_ctx *login_attempt;

	login_attempt = container_of(t, struct login_attempt_ctx, unlock_timer);
	login_guard_delete_attempt(login_attempt);
}

static struct login_attempt_ctx *login_guard_create_attempt(const char *username)
{
	struct login_attempt_ctx *attempt = calloc(1, sizeof(*attempt));
	if (attempt == NULL) {
		return NULL;
	}
	attempt->lock_timer.cb = login_guard_lock_timeout;
	attempt->unlock_timer.cb = login_guard_unlock_timeout;
	attempt->avl.key = strdup(username);
	if (attempt->avl.key == NULL) {
		free(attempt);
		return NULL;
	}
	avl_insert(&login_guard.login_attempts, &attempt->avl);
	return attempt;
}

static void login_guard_delete_attempt(struct login_attempt_ctx *attempt)
{
	avl_delete(&login_guard.login_attempts, &attempt->avl);
	uloop_timeout_cancel(&attempt->lock_timer);
	uloop_timeout_cancel(&attempt->unlock_timer);
	free((void *) attempt->avl.key);
	free(attempt);
}

bool login_guard_is_locked(const char *username)
{
	if (!login_guard.enabled)
		return false;

	struct login_attempt_ctx *attempt = login_guard_find_attempt(username);
	if (attempt != NULL) {
		if (attempt->is_locked) {
			int remaining_time = uloop_timeout_remaining(&attempt->unlock_timer);
			syslog(LOG_NOTICE, "user account '%s' locked for %d seconds", username, remaining_time / 1000);
			return true;
		} else {
			return false;
		}
	}
	return false;
}

void login_guard_add_attempt(const char *username) {
	if (!login_guard.enabled)
		return;

	struct login_attempt_ctx *attempt = login_guard_find_attempt(username);

	syslog(LOG_NOTICE, "failed login attempt for user: %s", username);
	if (attempt == NULL) { // create new attempt
		attempt = login_guard_create_attempt(username);
		if (attempt == NULL) {
			syslog(LOG_ERR, "memory allocation failed");
			return;
		}
		attempt->incorrect_attempts++;
		uloop_timeout_set(&attempt->lock_timer, login_guard.max_time_before_lock * 1000);
	} else { // increment attempt
		if (attempt->incorrect_attempts == 0) {
			uloop_timeout_set(&attempt->lock_timer, login_guard.max_time_before_lock * 1000);
		}
		attempt->incorrect_attempts++;
		if (attempt->incorrect_attempts >= login_guard.max_attempts) { // lock account
			attempt->is_locked = true;
			uloop_timeout_set(&attempt->unlock_timer, login_guard.lock_duration * 1000);
		}
	}
}

void login_guard_remove_attempt(const char *username)
{
	if (!login_guard.enabled)
		return;

	syslog(LOG_NOTICE, "successful login for user: %s", username);
	struct login_attempt_ctx *attempt = login_guard_find_attempt(username);
	if (attempt != NULL) {
		login_guard_delete_attempt(attempt);
	}
}

void login_guard_init()
{
	struct uci_context *uci = NULL;
	struct uci_package *p = NULL;
	struct uci_section *s;
	struct uci_element *e;
	struct uci_ptr ptr = { .package = "rpcd" };

	uci = uci_alloc_context();

	if (!uci) {
		goto out;
	}

	uci_load(uci, ptr.package, &p);

	if (!p)
		goto out;

	uci_foreach_element(&p->sections, e)
	{
		s = uci_to_section(e);

		if (strcmp(s->type, CONFIG_GUARD))
			continue;

		ptr.section = s->e.name;
		ptr.s = NULL;

		/* max_incorrrect_attempts parse */
		ptr.option = CONFIG_MAX_ATTEMPTS;
		ptr.o = NULL;

		if (uci_lookup_ptr(uci, &ptr, NULL, true))
			continue;

		if (ptr.o->type != UCI_TYPE_STRING)
			continue;

		int nr_attempts = atoi(ptr.o->v.string);
		if (nr_attempts <= 0)
			continue;

		login_guard.max_attempts = nr_attempts;

		/* max_time_before_locking parse */
		ptr.option = CONFIG_MAX_TIME_BEFORE_LOCK;
		ptr.o = NULL;

		if (uci_lookup_ptr(uci, &ptr, NULL, true))
			continue;

		if (ptr.o->type != UCI_TYPE_STRING)
			continue;

		int max_time = atoi(ptr.o->v.string);
		if (max_time <= 0)
			continue;

		login_guard.max_time_before_lock = max_time;

		/* lock_duration parse */
		ptr.option = CONFIG_LOCK_DURATION;
		ptr.o = NULL;

		if (uci_lookup_ptr(uci, &ptr, NULL, true))
			continue;

		if (ptr.o->type != UCI_TYPE_STRING)
			continue;

		int duration = atoi(ptr.o->v.string);
		if (duration  <= 0)
			continue;

		login_guard.lock_duration = duration;

		// all options present, enable login guard
		login_guard.enabled = true;
		avl_init(&login_guard.login_attempts, avl_strcmp, false, NULL);
		syslog(LOG_NOTICE, "login guard enabled");
		break;
	}

out:
	if (uci)
		uci_free_context(uci);

	return;
}
