/*
 * granular uci access control
 *
 * Copyright (C) 2019 IOPSYS
 *
 * Author: Jakob Olsson <jakob.olsson@iopsys.eu>
 * Author: Matija Amidzic <matija.amidzic@sartura.hr>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libubox/avl.h>
#include <libubox/avl-cmp.h>
#include <libubox/blobmsg.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include <rpcd/session.h>
#include <rpcd/uci_granular.h>

#include <syslog.h>

static struct uci_context *_uci_context;

struct uci_granular_s {
	struct avl_node avl;

	char username[GRANULAR_MAX_STRING_LENGTH];
	struct match_definition_s match_definitions[GRANULAR_MAX_MATCH_DEFINITIONS];
	int match_definitions_count;
};

static struct uci_granular_s *uci_granular_find_config(struct rpc_session *ses, const char *config_name, const char *username, struct uci_granular_ctx_s *granular_ctx);
//static int uci_granular_parse_match_definition(struct blob_attr *acl_definition, struct match_definition_s *match_definition);
static bool rpc_uci_granular_match_by_option(const char *config_name, const char *section_name, const char *section_type, struct match_option_s *option);
static void uci_granular_free_config(struct rpc_session *ses, struct uci_granular_s *config);
static bool rpc_uci_granular_section_is_match(struct match_definition_s *match_definition, struct uci_granular_query_s *query);

int rpc_uci_granular_init(struct rpc_session *ses)
{
	//syslog(LOG_DEBUG, "rpc_uci_granular_init");
	avl_init(&ses->granular_configs, avl_strcmp, false, NULL);
	if (!_uci_context)
		_uci_context = uci_alloc_context();
	if (!_uci_context) {
		//syslog(LOG_DEBUG, "rpc_uci_granular_init failed alloc of uci context");
		return -1;
	}

	return 0;
}

int rpc_uci_granular_destroy(void)
{
	if (_uci_context)
		uci_free_context(_uci_context);

	return 0;
}


int rpc_uci_granular_setup(struct rpc_session *ses, struct blob_attr *acl_top, const char *username)
{
	/*
	 *	"uci_granular": {
	 *		"<config>" : [{
	 *			"match:" {
	 *				"<match type>": "<match string>",
	 *				...
	 *			},
	 *			"option:" [
	 *				<option_name> or '*' for all
	 *				...
	 *			]
	 *		},
	 *		...
	 *		]
	 *	}
	 */

	//syslog(LOG_DEBUG, "rpc_uci_setup_granular enter");

	struct blob_attr *acl_config = {0};
	int rem1 = 0;
	bool new = false;

	blobmsg_for_each_attr(acl_config, acl_top, rem1) {
		if (blobmsg_type(acl_config) != BLOBMSG_TYPE_ARRAY) {
			continue;
		}
		struct uci_granular_s *granular_config = NULL;
		struct uci_granular_s *temp = NULL;

		avl_for_each_element(&ses->granular_configs, temp, avl) {
			if (strcmp((char *)temp->avl.key, blobmsg_name(acl_config)) == 0) {
				granular_config = temp;
				break;
			}
		}

		if (!granular_config) {
			granular_config = calloc(1, sizeof(*granular_config));
			if (!granular_config)
				return -1;

			granular_config->avl.key = strdup(blobmsg_name(acl_config));
			strncpy(granular_config->username, username, GRANULAR_MAX_STRING_LENGTH);
			if (granular_config->avl.key == NULL) {
				free(granular_config);
				//syslog(LOG_DEBUG, "rpc_uci_setup_granular failed allocation of avl key");
				continue;
			}
			new = true;
		}

		struct blob_attr *acl_match_definition = {0};
		int rem2 = 0;
		blobmsg_for_each_attr(acl_match_definition, acl_config, rem2) {
			if (blobmsg_type(acl_match_definition) != BLOBMSG_TYPE_TABLE) {
				continue;
			}

			uci_granular_parse_match_definition(acl_match_definition, &granular_config->match_definitions[granular_config->match_definitions_count]);
			granular_config->match_definitions_count++;
			if (granular_config->match_definitions_count >= GRANULAR_MAX_MATCH_DEFINITIONS) {
				//syslog(LOG_DEBUG, "rpc_uci_setup_granular exceeded possible number of match definitions");
				break;
			}
		}

		if (new)
			avl_insert(&(ses->granular_configs), &granular_config->avl);
	}

	return 0;
}


int uci_granular_parse_match_definition(struct blob_attr *acl_match_definition, struct match_definition_s *match_definition)
{
	struct blob_attr *acl_obj = {0};
	int rem3 = 0;
	blobmsg_for_each_attr(acl_obj, acl_match_definition, rem3)
	{
		if (strcmp(blobmsg_name(acl_obj), "match") == 0) {
			if (blobmsg_type(acl_obj) != BLOBMSG_TYPE_TABLE) {
				continue;
			}
			struct blob_attr *acl_match = {0};
			int rem4 = 0;
			blobmsg_for_each_attr(acl_match, acl_obj, rem4)
			{
				if (blobmsg_type(acl_match) != BLOBMSG_TYPE_STRING) {
					continue;
				}
				//syslog(LOG_DEBUG, "rpc_uci_setup_granular match: %s:%s", blobmsg_name(acl_match), (char *) blobmsg_data(acl_match));

				if (strcmp(blobmsg_name(acl_match), ".type") == 0) {
					strncpy(match_definition->match_section_type, (char *) blobmsg_data(acl_match), GRANULAR_MAX_STRING_LENGTH - 1);
					continue;
				} else if (strcmp(blobmsg_name(acl_match), ".name") == 0) {
					strncpy(match_definition->match_section_name, (char *) blobmsg_data(acl_match), GRANULAR_MAX_STRING_LENGTH - 1);
					continue;
				} else if (match_definition->match_options_count < GRANULAR_MAX_MATCH_OPTIONS) {
					strncpy(match_definition->match_options[match_definition->match_options_count].name, blobmsg_name(acl_match), GRANULAR_MAX_STRING_LENGTH - 1);
					strncpy(match_definition->match_options[match_definition->match_options_count].value, (char *) blobmsg_data(acl_match), GRANULAR_MAX_STRING_LENGTH - 1);
					match_definition->match_options_count++;
					continue;
				}
			}
		} else if (strcmp(blobmsg_name(acl_obj), "option") == 0) {
			if (blobmsg_type(acl_obj) != BLOBMSG_TYPE_ARRAY) {
				continue;
			}
			struct blob_attr *acl_option = {0};
			int rem4 = 0;
			int i = match_definition->allowed_options_count;
			char **allowed_options;

			blobmsg_for_each_attr(acl_option, acl_obj, rem4)
			{
				if (blobmsg_type(acl_option) != BLOBMSG_TYPE_STRING) {
					continue;
				}
				match_definition->allowed_options_count++;
			}

			if (match_definition->allowed_options_count >= GRANULAR_MAX_ALLOWED_OPTIONS) {
				match_definition->allowed_options_count = GRANULAR_MAX_MATCH_OPTIONS;
			}

			allowed_options = realloc(match_definition->allowed_options, match_definition->allowed_options_count * sizeof(char *));
			if (!allowed_options)
				return -1;

			match_definition->allowed_options = allowed_options;

			rem4 = 0;
			blobmsg_for_each_attr(acl_option, acl_obj, rem4)
			{
				if (blobmsg_type(acl_option) != BLOBMSG_TYPE_STRING) {
					continue;
				}
				//syslog(LOG_DEBUG, "rpc_uci_setup_granular allowed option name: %s", (char *) blobmsg_data(acl_option));

				if (match_definition->allowed_options_count >= GRANULAR_MAX_ALLOWED_OPTIONS) {
					//syslog(LOG_DEBUG, "rpc_uci_setup_granular  exceeded number of options");
					continue;
				}
				match_definition->allowed_options[i++] = strndup((char *) blobmsg_data(acl_option), GRANULAR_MAX_STRING_LENGTH);
			}
		}
	}

	return 0;
}

void rpc_uci_granular_get_allowed(struct rpc_session *ses, struct uci_granular_query_s *query, struct uci_granular_ctx_s *granular_ctx)
{
	// find relevant uci_granular_s data structure
	struct uci_granular_s *config = NULL;
	int j = 0;

	memset(granular_ctx->match_definitions, 0, sizeof(granular_ctx->match_definitions));
	granular_ctx->match_definitions_count = 0;

	config = uci_granular_find_config(ses, query->config_name, query->username, granular_ctx);
	// allow if nothing is specified for this config
	if (!config) {
		granular_ctx->match_definitions_count = -1;
		return;
	}

	for (int i = 0; i < config->match_definitions_count; i++) {
		struct match_definition_s *match_definition = &config->match_definitions[i];

		// if match definition has section type defined, check query
		if (!rpc_uci_granular_section_is_match(match_definition, query))
			continue;
		granular_ctx->match_definitions[j++] = match_definition;
		granular_ctx->match_definitions_count++;
	}

	return;
}

static bool rpc_uci_granular_section_is_match(
	struct match_definition_s *match_definition,
	struct uci_granular_query_s *query)
{
	if (query->section_type || query->section_name)
	{
		if (strcmp(match_definition->match_section_type, "") != 0)
		{
			// if query doesn't have section type defined or it isn't the same as in match definition, match is false
			if (query->section_type == NULL || strcmp(query->section_type, match_definition->match_section_type) != 0)
				return false;
		}

		// if match definition has section name defined, check query
		if (strcmp(match_definition->match_section_name, "") != 0)
		{
			// if query doesn't have section name defined or it isn't the same as in match definition, match is false
			if (query->section_name == NULL || strcmp(query->section_name, match_definition->match_section_name) != 0)
				return false;
		}
	}

	// match by options
	bool is_match = true;
	for (int j = 0; j < match_definition->match_options_count; j++) {
		syslog(LOG_DEBUG, "rpc_uci_granular_match_section checking %d. match option: %s:%s", j, match_definition->match_options[j].name, match_definition->match_options[j].value);
		// find out if the option exists and it's value is a match in the given config/section
		is_match = rpc_uci_granular_match_by_option(
			query->config_name, query->section_name, query->section_type, &match_definition->match_options[j]);
		if (!is_match) // this is not the match definition we are looking for
			break;
	}

	return is_match;
}

static bool
rpc_uci_granular_match_by_option(const char *config_name, const char *section_name, const char *section_type, struct match_option_s *option)
{
	if (!_uci_context) {
		return true;
	}

	if (section_name && strcmp(section_name, "") != 0) {
		char uci_string[GRANULAR_MAX_STRING_LENGTH * 3] = {0}; // times 3 to allow config, section and option
		size_t uci_string_max_length = GRANULAR_MAX_STRING_LENGTH * 3;
		struct uci_ptr lookup = {0};

		snprintf(uci_string, uci_string_max_length, "%s.%s.%s", config_name, section_name, option->name);
		//syslog(LOG_DEBUG, "rpc_uci_granular_match_by_option uci_string: %s", uci_string);

		lookup.flags |= UCI_LOOKUP_EXTENDED;
		int rc = uci_lookup_ptr(_uci_context, &lookup, uci_string, true);
		if (rc == 0 && lookup.flags & UCI_LOOKUP_COMPLETE) {
			uci_unload(_uci_context, lookup.p);

			// found relevant option, now need to check if the value matches
			if (lookup.o->type == UCI_TYPE_STRING && strcmp(option->value, lookup.o->v.string) == 0) {
				//syslog(LOG_DEBUG, "rpc_uci_granular_match_by_option option match, option value: %s", option->value);
				return true;
			} else {
				//syslog(LOG_DEBUG, "rpc_uci_granular_match_by_option option didn't match, option value: %s", option->value);
				return false;
			}
		} else {
			//syslog(LOG_DEBUG, "rpc_uci_granular_match_by_option didn't find uci string: %s", uci_string);
			return false;
		}
	} else {
		struct uci_package *p = NULL;
		struct uci_section *s;
		struct uci_element *e;
		struct uci_ptr ptr = {.package = config_name};
		bool access = false;

		//syslog(LOG_DEBUG, "rpc_uci_granular_match_by_option searching all sections");
		uci_load(_uci_context, ptr.package, &p);
		if (!p) {
			// this shouldn't ever happen
			//syslog(LOG_DEBUG, "rpc_uci_granular_match_by_option couldn't find package: %s", config_name);
			return false;
		}

		uci_foreach_element(&p->sections, e) {
			s = uci_to_section(e);
			if (section_type && strncmp(section_type, "", 2) != 0 &&
					 strncmp(s->type, section_type, 64) != 0)
				continue;

			const char *found_option_value = uci_lookup_option_string(_uci_context, s, option->name);
			if (found_option_value != NULL && strcmp(option->value, found_option_value) == 0) {
				access = true;
				break;
			}
		}

		uci_unload(_uci_context, p);
		return access;
	}
}

static struct uci_granular_s *uci_granular_find_config(struct rpc_session *ses,
	const char *config_name, const char *username,
	struct uci_granular_ctx_s *granular_ctx)
{
	//syslog(LOG_DEBUG, "rpc_uci_granular_find_config config_name:%s username:%s", config_name, username);
	struct uci_granular_s *config = NULL;
	struct uci_granular_s *temp = NULL;

	granular_ctx->state = RPC_CA_ALLOWED;

	avl_for_each_element(&ses->granular_configs, temp, avl)
	{
		granular_ctx->state = RPC_CA_RESTRICTED;

		//syslog(LOG_DEBUG, "rpc_uci_granular_find_config checking config config_name:%s username:%s", (char *) temp->avl.key, temp->username);
		if (strcmp((char *) temp->avl.key, config_name) == 0) {
			config = temp;
			//syslog(LOG_DEBUG, "rpc_uci_granular_find_config found config %s", username);
			// we found the config, break
			break;
		}
	}
	return config;
}

static void uci_granular_free_config(struct rpc_session *ses, struct uci_granular_s *config)
{
	avl_delete(&ses->granular_configs, &config->avl);
	for (int i = 0; i < config->match_definitions_count; i++) {
		struct match_definition_s *match_definition = &config->match_definitions[i];

		for (int j = 0; j < match_definition->allowed_options_count; j++) {
			free(match_definition->allowed_options[j]);
		}
		free(match_definition->allowed_options);
	}
	free((void *)config->avl.key);
	free(config);
}

void uci_granular_exit(struct rpc_session *ses)
{
	struct uci_granular_s *config, *nconfig;

	avl_for_each_element_safe(&ses->granular_configs, config, avl, nconfig)
	{
		uci_granular_free_config(ses, config);
	}
}
